"""Data models."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any

import voluptuous as vol  # type:ignore[import]


@dataclass
class Timestamp:
    """Timestamp."""

    SCHEMA = vol.Schema(
        {
            vol.Optional("Authoritative"): bool,
            vol.Optional("Seconds"): int,
            vol.Optional("Ticks"): int,
        }
    )

    authoritative: bool | None = None
    seconds: int | None = None
    ticks: int | None = None

    def as_json(self) -> dict:
        """Serialize to JSON."""
        result: dict[str, Any] = {}
        if self.authoritative is not None:
            result["Authoritative"] = self.authoritative
        if self.seconds is not None:
            result["Seconds"] = self.seconds
        if self.ticks is not None:
            result["Ticks"] = self.ticks
        return result

    @classmethod
    def from_json(cls, json_data: Any) -> Timestamp:
        """Deserialize from JSON."""
        cls.SCHEMA(json_data)
        return cls(
            json_data.get("Authoritative"),
            json_data.get("Seconds"),
            json_data.get("Ticks"),
        )


@dataclass
class SecurityPolicy:  # pylint: disable=too-many-instance-attributes
    """Security policy."""

    SCHEMA = vol.Schema(
        {
            vol.Optional("AutonomousEnrollment"): bool,
            vol.Optional("CommercialCommissioning"): bool,
            vol.Optional("ExternalCommissioning"): bool,
            vol.Optional("NativeCommissioning"): bool,
            vol.Optional("NetworkKeyProvisioning"): bool,
            vol.Optional("NonCcmRouters"): bool,
            vol.Optional("ObtainNetworkKey"): bool,
            vol.Optional("RotationTime"): int,
            vol.Optional("Routers"): bool,
            vol.Optional("TobleLink"): bool,
        }
    )

    autonomous_enrollment: bool | None = None
    commercial_commissioning: bool | None = None
    external_commissioning: bool | None = None
    native_commissioning: bool | None = None
    network_key_provisioning: bool | None = None
    non_ccm_routers: bool | None = None
    obtain_network_key: bool | None = None
    rotation_time: int | None = None
    routers: bool | None = None
    to_ble_link: bool | None = None

    def as_json(self) -> dict:
        """Serialize to JSON."""
        result: dict[str, Any] = {}
        if self.autonomous_enrollment is not None:
            result["AutonomousEnrollment"] = self.autonomous_enrollment
        if self.commercial_commissioning is not None:
            result["CommercialCommissioning"] = self.commercial_commissioning
        if self.external_commissioning is not None:
            result["ExternalCommissioning"] = self.external_commissioning
        if self.native_commissioning is not None:
            result["NativeCommissioning"] = self.native_commissioning
        if self.network_key_provisioning is not None:
            result["NetworkKeyProvisioning"] = self.network_key_provisioning
        if self.non_ccm_routers is not None:
            result["NonCcmRouters"] = self.non_ccm_routers
        if self.obtain_network_key is not None:
            result["ObtainNetworkKey"] = self.obtain_network_key
        if self.rotation_time is not None:
            result["RotationTime"] = self.rotation_time
        if self.routers is not None:
            result["Routers"] = self.routers
        if self.to_ble_link is not None:
            result["TobleLink"] = self.to_ble_link
        return result

    @classmethod
    def from_json(cls, json_data: Any) -> SecurityPolicy:
        """Deserialize from JSON."""
        cls.SCHEMA(json_data)
        return cls(
            json_data.get("AutonomousEnrollment"),
            json_data.get("CommercialCommissioning"),
            json_data.get("ExternalCommissioning"),
            json_data.get("NativeCommissioning"),
            json_data.get("NetworkKeyProvisioning"),
            json_data.get("NonCcmRouters"),
            json_data.get("ObtainNetworkKey"),
            json_data.get("RotationTime"),
            json_data.get("Routers"),
            json_data.get("TobleLink"),
        )


@dataclass
class ActiveDataSet:  # pylint: disable=too-many-instance-attributes
    """Operational dataset."""

    SCHEMA = vol.Schema(
        {
            vol.Optional("ActiveTimestamp"): dict,
            vol.Optional("ChannelMask"): int,
            vol.Optional("Channel"): int,
            vol.Optional("ExtPanId"): str,
            vol.Optional("MeshLocalPrefix"): str,
            vol.Optional("NetworkKey"): str,
            vol.Optional("NetworkName"): str,
            vol.Optional("PanId"): int,
            vol.Optional("PSKc"): str,
            vol.Optional("SecurityPolicy"): dict,
        }
    )

    active_timestamp: Timestamp | None = None
    channel_mask: int | None = None
    channel: int | None = None
    extended_pan_id: str | None = None
    mesh_local_prefix: str | None = None
    network_key: str | None = None
    network_name: str | None = None
    pan_id: int | None = None
    psk_c: str | None = None
    security_policy: SecurityPolicy | None = None

    def as_json(self) -> dict:
        """Serialize to JSON."""
        result: dict[str, Any] = {}
        if self.active_timestamp is not None:
            result["ActiveTimestamp"] = self.active_timestamp.as_json()
        if self.channel_mask is not None:
            result["ChannelMask"] = self.channel_mask
        if self.channel is not None:
            result["Channel"] = self.channel
        if self.extended_pan_id is not None:
            result["ExtPanId"] = self.extended_pan_id
        if self.mesh_local_prefix is not None:
            result["MeshLocalPrefix"] = self.mesh_local_prefix
        if self.network_key is not None:
            result["NetworkKey"] = self.network_key
        if self.network_name is not None:
            result["NetworkName"] = self.network_name
        if self.pan_id is not None:
            result["PanId"] = self.pan_id
        if self.psk_c is not None:
            result["PSKc"] = self.psk_c
        if self.security_policy is not None:
            result["SecurityPolicy"] = self.security_policy.as_json()
        return result

    @classmethod
    def from_json(cls, json_data: Any) -> ActiveDataSet:
        """Deserialize from JSON."""
        cls.SCHEMA(json_data)
        active_timestamp = None
        security_policy = None
        if "ActiveTimestamp" in json_data:
            active_timestamp = Timestamp.from_json(json_data["ActiveTimestamp"])
        if "SecurityPolicy" in json_data:
            security_policy = SecurityPolicy.from_json(json_data["SecurityPolicy"])

        return ActiveDataSet(
            active_timestamp,
            json_data.get("ChannelMask"),
            json_data.get("Channel"),
            json_data.get("ExtPanId"),
            json_data.get("MeshLocalPrefix"),
            json_data.get("NetworkKey"),
            json_data.get("NetworkName"),
            json_data.get("PanId"),
            json_data.get("PSKc"),
            security_policy,
        )


@dataclass
class PendingDataSet:  # pylint: disable=too-many-instance-attributes
    """Operational dataset."""

    SCHEMA = vol.Schema(
        {
            vol.Optional("ActiveDataset"): dict,
            vol.Optional("Delay"): int,
            vol.Optional("PendingTimestamp"): dict,
        }
    )

    active_dataset: ActiveDataSet | None = None
    delay: int | None = None
    pending_timestamp: Timestamp | None = None

    def as_json(self) -> dict:
        """Serialize to JSON."""
        result: dict[str, Any] = {}
        if self.active_dataset is not None:
            result["ActiveDataset"] = self.active_dataset.as_json()
        if self.delay is not None:
            result["Delay"] = self.delay
        if self.pending_timestamp is not None:
            result["PendingTimestamp"] = self.pending_timestamp.as_json()
        return result

    @classmethod
    def from_json(cls, json_data: Any) -> PendingDataSet:
        """Deserialize from JSON."""
        cls.SCHEMA(json_data)
        active_dataset = None
        pending_timestamp = None
        if "ActiveDataset" in json_data:
            active_dataset = ActiveDataSet.from_json(json_data["ActiveDataset"])
        if "PendingTimestamp" in json_data:
            pending_timestamp = Timestamp.from_json(json_data["PendingTimestamp"])

        return PendingDataSet(
            active_dataset,
            json_data.get("Delay"),
            pending_timestamp,
        )



"""Calculate Thread PSKc.

Based on https://github.com/openthread/ot-br-posix/blob/main/src/utils/pskc.cpp
"""

import struct

from cryptography.hazmat.primitives import cmac
from cryptography.hazmat.primitives.ciphers import algorithms

AES_128_KEY_LEN = 16
ITERATION_COUNTS = 16384
BLKSIZE = 16
SALT_PREFIX = "Thread".encode()


def _derive_key(passphrase: str) -> bytes:
    """Derive key from passphrase according to RFC 4615."""
    passphrase_bytes = passphrase.encode()
    if len(passphrase_bytes) == AES_128_KEY_LEN:
        return passphrase_bytes
    c = cmac.CMAC(algorithms.AES128(b"\0" * AES_128_KEY_LEN))
    c.update(passphrase_bytes)
    return c.finalize()


def compute_pskc(ext_pan_id: bytes, network_name: str, passphrase: str) -> bytes:
    """Compute Thread PSKc."""
    salt = SALT_PREFIX + ext_pan_id + network_name.encode()
    key = _derive_key(passphrase)

    block_counter = 1
    prf_input = salt + struct.pack("!L", block_counter)

    # Calculate U_1
    c = cmac.CMAC(algorithms.AES128(key))
    c.update(prf_input)
    prf_output = c.finalize()
    pskc = bytearray(prf_output)

    for _ in range(ITERATION_COUNTS - 1):
        prf_input = prf_output

        # Calculate U_i
        c = cmac.CMAC(algorithms.AES128(key))
        c.update(prf_input)
        prf_output = c.finalize()

        # xor
        for i in range(BLKSIZE):
            pskc[i] ^= prf_output[i]

    return pskc



"""Test fixtures."""

from collections.abc import Generator

import pytest

from tests.test_util.aiohttp import AiohttpClientMocker, mock_aiohttp_client


@pytest.fixture
def aioclient_mock() -> Generator[AiohttpClientMocker, None, None]:
    """Fixture to mock aioclient calls."""
    with mock_aiohttp_client() as mock_session:
        yield mock_session



"""Test decoding fields in _meshcop._udp.local. services."""

import pytest

from python_otbr_api.mdns import (
    Availability,
    ConnectionMode,
    StateBitmap,
    ThreadInterfaceStatus,
)


@pytest.mark.parametrize(
    "encoded, decoded",
    [
        (
            b"\x00\x00\x01\xb1",
            StateBitmap(
                connection_mode=ConnectionMode.PSKC,
                thread_interface_status=ThreadInterfaceStatus.ACTIVE,
                availability=Availability.HIGH,
                is_active=True,
                is_primary=True,
            ),
        ),
        (
            b"\x00\x00\x00!",
            StateBitmap(
                connection_mode=ConnectionMode.PSKC,
                thread_interface_status=ThreadInterfaceStatus.NOT_INITIALIZED,
                availability=Availability.HIGH,
                is_active=False,
                is_primary=False,
            ),
        ),
    ],
)
def test_decode_state_bitmap(encoded, decoded) -> None:
    """Test the TLV parser."""
    assert StateBitmap.from_bytes(encoded) == decoded


@pytest.mark.parametrize(
    "encoded, error",
    [
        # Input not bytes
        ("blah", TypeError),
        # Wrong length
        (b"\x00\x01\xb1", ValueError),
        # Padding not zeroed
        (b"\xff\x00\x01\xb1", ValueError),
        # Invalid ConnectionMode
        (b"\x00\x00\x01\xb5", ValueError),
        # Invalid ThreadInterfaceStatus
        (b"\x00\x00\x01\xb9", ValueError),
        # Invalid Availability
        (b"\x00\x00\x01\xf1", ValueError),
    ],
)
def test_decode_state_bitmap_error(encoded, error) -> None:
    """Test the TLV parser."""
    with pytest.raises(error):
        StateBitmap.from_bytes(encoded)



"""Test the Thread TLV parser."""

import pytest

from python_otbr_api.tlv_parser import (
    Timestamp,
    Channel,
    MeshcopTLVItem,
    MeshcopTLVType,
    NetworkName,
    TLVError,
    encode_tlv,
    parse_tlv,
)


def test_encode_tlv() -> None:
    """Test the TLV parser."""
    dataset = {
        MeshcopTLVType.ACTIVETIMESTAMP: MeshcopTLVItem(
            MeshcopTLVType.ACTIVETIMESTAMP, bytes.fromhex("0000000000010000")
        ),
        MeshcopTLVType.CHANNEL: MeshcopTLVItem(
            MeshcopTLVType.CHANNEL, bytes.fromhex("00000f")
        ),
        MeshcopTLVType.CHANNELMASK: MeshcopTLVItem(
            MeshcopTLVType.CHANNELMASK, bytes.fromhex("0004001fffe0")
        ),
        MeshcopTLVType.EXTPANID: MeshcopTLVItem(
            MeshcopTLVType.EXTPANID, bytes.fromhex("1111111122222222")
        ),
        MeshcopTLVType.MESHLOCALPREFIX: MeshcopTLVItem(
            MeshcopTLVType.MESHLOCALPREFIX, bytes.fromhex("fdad70bfe5aa15dd")
        ),
        MeshcopTLVType.NETWORKKEY: MeshcopTLVItem(
            MeshcopTLVType.NETWORKKEY, bytes.fromhex("00112233445566778899aabbccddeeff")
        ),
        MeshcopTLVType.NETWORKNAME: NetworkName(
            MeshcopTLVType.NETWORKNAME, "OpenThreadDemo".encode()
        ),
        MeshcopTLVType.PANID: MeshcopTLVItem(
            MeshcopTLVType.PANID, bytes.fromhex("1234")
        ),
        MeshcopTLVType.PSKC: MeshcopTLVItem(
            MeshcopTLVType.PSKC, bytes.fromhex("445f2b5ca6f2a93a55ce570a70efeecb")
        ),
        MeshcopTLVType.SECURITYPOLICY: MeshcopTLVItem(
            MeshcopTLVType.SECURITYPOLICY, bytes.fromhex("02a0f7f8")
        ),
    }
    dataset_tlv = encode_tlv(dataset)
    assert (
        dataset_tlv
        == (
            "0E080000000000010000000300000F35060004001FFFE0020811111111222222220708FDAD"
            "70BFE5AA15DD051000112233445566778899AABBCCDDEEFF030E4F70656E54687265616444"
            "656D6F010212340410445F2B5CA6F2A93A55CE570A70EFEECB0C0402A0F7F8"
        ).lower()
    )


def test_parse_tlv() -> None:
    """Test the TLV parser."""
    dataset_tlv = (
        "0E080000000000010000000300000F35060004001FFFE0020811111111222222220708FDAD70BF"
        "E5AA15DD051000112233445566778899AABBCCDDEEFF030E4F70656E54687265616444656D6F01"
        "0212340410445F2B5CA6F2A93A55CE570A70EFEECB0C0402A0F7F8"
    )
    dataset = parse_tlv(dataset_tlv)
    assert dataset == {
        MeshcopTLVType.CHANNEL: Channel(
            MeshcopTLVType.CHANNEL, bytes.fromhex("00000f")
        ),
        MeshcopTLVType.PANID: MeshcopTLVItem(
            MeshcopTLVType.PANID, bytes.fromhex("1234")
        ),
        MeshcopTLVType.EXTPANID: MeshcopTLVItem(
            MeshcopTLVType.EXTPANID, bytes.fromhex("1111111122222222")
        ),
        MeshcopTLVType.NETWORKNAME: NetworkName(
            MeshcopTLVType.NETWORKNAME, "OpenThreadDemo".encode()
        ),
        MeshcopTLVType.PSKC: MeshcopTLVItem(
            MeshcopTLVType.PSKC, bytes.fromhex("445f2b5ca6f2a93a55ce570a70efeecb")
        ),
        MeshcopTLVType.NETWORKKEY: MeshcopTLVItem(
            MeshcopTLVType.NETWORKKEY, bytes.fromhex("00112233445566778899aabbccddeeff")
        ),
        MeshcopTLVType.MESHLOCALPREFIX: MeshcopTLVItem(
            MeshcopTLVType.MESHLOCALPREFIX, bytes.fromhex("fdad70bfe5aa15dd")
        ),
        MeshcopTLVType.SECURITYPOLICY: MeshcopTLVItem(
            MeshcopTLVType.SECURITYPOLICY, bytes.fromhex("02a0f7f8")
        ),
        MeshcopTLVType.ACTIVETIMESTAMP: Timestamp(
            MeshcopTLVType.ACTIVETIMESTAMP, bytes.fromhex("0000000000010000")
        ),
        MeshcopTLVType.CHANNELMASK: MeshcopTLVItem(
            MeshcopTLVType.CHANNELMASK, bytes.fromhex("0004001fffe0")
        ),
    }


def test_parse_tlv_apple() -> None:
    """Test the TLV parser from a (truncated) dataset from an Apple BR."""
    dataset_tlv = (
        "0e08000065901a07000000030000194a0300000f35060004001fffc003104d79486f6d65313233"
        "31323331323334"
    )
    dataset = parse_tlv(dataset_tlv)
    assert dataset == {
        MeshcopTLVType.ACTIVETIMESTAMP: Timestamp(
            MeshcopTLVType.ACTIVETIMESTAMP, bytes.fromhex("000065901a070000")
        ),
        MeshcopTLVType.CHANNEL: Channel(
            MeshcopTLVType.CHANNEL, bytes.fromhex("000019")
        ),
        MeshcopTLVType.APPLE_TAG_UNKNOWN: MeshcopTLVItem(
            MeshcopTLVType.APPLE_TAG_UNKNOWN, bytes.fromhex("00000f")
        ),
        MeshcopTLVType.CHANNELMASK: MeshcopTLVItem(
            MeshcopTLVType.CHANNELMASK, bytes.fromhex("0004001fffc0")
        ),
        MeshcopTLVType.NETWORKNAME: NetworkName(
            MeshcopTLVType.NETWORKNAME, "MyHome1231231234".encode()
        ),
    }


@pytest.mark.parametrize(
    "tlv, error, msg",
    (
        (
            "killevippen",
            TLVError,
            "invalid tlvs",
        ),
        (
            "FF",
            TLVError,
            "unknown type 255",
        ),
        (
            "030E4F70656E54687265616444656D",
            TLVError,
            "expected 14 bytes for NETWORKNAME, got 13",
        ),
        (
            "030E4F70656E54687265616444656DFF",
            TLVError,
            "invalid network name '4f70656e54687265616444656dff'",
        ),
    ),
)
def test_parse_tlv_error(tlv, error, msg) -> None:
    """Test the TLV parser error handling."""
    with pytest.raises(error, match=msg):
        parse_tlv(tlv)



"""API to interact with the Open Thread Border Router REST API."""

from __future__ import annotations
from http import HTTPStatus
import json

import aiohttp
import voluptuous as vol  # type:ignore[import]

from .models import ActiveDataSet, PendingDataSet, Timestamp

# 5 minutes as recommended by
# https://github.com/openthread/openthread/discussions/8567#discussioncomment-4468920
PENDING_DATASET_DELAY_TIMER = 5 * 60 * 1000


class OTBRError(Exception):
    """Raised on error."""


class FactoryResetNotSupportedError(OTBRError):
    """Raised when attempting to factory reset a router which does not support it."""


class GetBorderAgentIdNotSupportedError(OTBRError):
    """Raised when attempting to get the agent ID if the router does not support it."""


class ThreadNetworkActiveError(OTBRError):
    """Raised on attempts to modify the active dataset when thread network is active."""


class OTBR:  # pylint: disable=too-few-public-methods
    """Class to interact with the Open Thread Border Router REST API."""

    def __init__(
        self, url: str, session: aiohttp.ClientSession, timeout: int = 10
    ) -> None:
        """Initialize."""
        self._session = session
        self._url = url
        self._timeout = timeout

    async def factory_reset(self) -> None:
        """Factory reset the router."""

        response = await self._session.delete(
            f"{self._url}/node",
            timeout=aiohttp.ClientTimeout(total=10),
        )

        if response.status == HTTPStatus.METHOD_NOT_ALLOWED:
            raise FactoryResetNotSupportedError

        if response.status != HTTPStatus.OK:
            raise OTBRError(f"unexpected http status {response.status}")

    async def get_border_agent_id(self) -> bytes:
        """Get the border agent ID."""
        response = await self._session.get(
            f"{self._url}/node/ba-id",
            timeout=aiohttp.ClientTimeout(total=self._timeout),
        )

        if response.status == HTTPStatus.NOT_FOUND:
            raise GetBorderAgentIdNotSupportedError

        if response.status != HTTPStatus.OK:
            raise OTBRError(f"unexpected http status {response.status}")

        try:
            return bytes.fromhex(await response.json())
        except ValueError as exc:
            raise OTBRError("unexpected API response") from exc

    async def set_enabled(self, enabled: bool) -> None:
        """Enable or disable the router."""

        response = await self._session.put(
            f"{self._url}/node/state",
            json="enable" if enabled else "disable",
            timeout=aiohttp.ClientTimeout(total=10),
        )

        if response.status != HTTPStatus.OK:
            raise OTBRError(f"unexpected http status {response.status}")

    async def get_active_dataset(self) -> ActiveDataSet | None:
        """Get current active operational dataset.

        Returns None if there is no active operational dataset.
        Raises if the http status is 400 or higher or if the response is invalid.
        """
        response = await self._session.get(
            f"{self._url}/node/dataset/active",
            timeout=aiohttp.ClientTimeout(total=self._timeout),
        )

        if response.status == HTTPStatus.NO_CONTENT:
            return None

        if response.status != HTTPStatus.OK:
            raise OTBRError(f"unexpected http status {response.status}")

        try:
            return ActiveDataSet.from_json(await response.json())
        except (json.JSONDecodeError, vol.Error) as exc:
            raise OTBRError("unexpected API response") from exc

    async def get_active_dataset_tlvs(self) -> bytes | None:
        """Get current active operational dataset in TLVS format, or None.

        Returns None if there is no active operational dataset.
        Raises if the http status is 400 or higher or if the response is invalid.
        """
        response = await self._session.get(
            f"{self._url}/node/dataset/active",
            headers={"Accept": "text/plain"},
            timeout=aiohttp.ClientTimeout(total=self._timeout),
        )

        if response.status == HTTPStatus.NO_CONTENT:
            return None

        if response.status != HTTPStatus.OK:
            raise OTBRError(f"unexpected http status {response.status}")

        try:
            return bytes.fromhex(await response.text("ASCII"))
        except ValueError as exc:
            raise OTBRError("unexpected API response") from exc

    async def get_pending_dataset_tlvs(self) -> bytes | None:
        """Get current pending operational dataset in TLVS format, or None.

        Returns None if there is no pending operational dataset.
        Raises if the http status is 400 or higher or if the response is invalid.
        """
        response = await self._session.get(
            f"{self._url}/node/dataset/pending",
            headers={"Accept": "text/plain"},
            timeout=aiohttp.ClientTimeout(total=self._timeout),
        )

        if response.status == HTTPStatus.NO_CONTENT:
            return None

        if response.status != HTTPStatus.OK:
            raise OTBRError(f"unexpected http status {response.status}")

        try:
            return bytes.fromhex(await response.text("ASCII"))
        except ValueError as exc:
            raise OTBRError("unexpected API response") from exc

    async def create_active_dataset(self, dataset: ActiveDataSet) -> None:
        """Create active operational dataset.

        The passed in ActiveDataSet does not need to be fully populated, any fields
        not set will be automatically set by the open thread border router.
        Raises if the http status is 400 or higher or if the response is invalid.
        """
        response = await self._session.put(
            f"{self._url}/node/dataset/active",
            json=dataset.as_json(),
            timeout=aiohttp.ClientTimeout(total=self._timeout),
        )

        if response.status == HTTPStatus.CONFLICT:
            raise ThreadNetworkActiveError
        if response.status not in (HTTPStatus.CREATED, HTTPStatus.OK):
            raise OTBRError(f"unexpected http status {response.status}")

    async def delete_active_dataset(self) -> None:
        """Delete active operational dataset."""
        response = await self._session.delete(
            f"{self._url}/node/dataset/active",
            timeout=aiohttp.ClientTimeout(total=self._timeout),
        )

        if response.status == HTTPStatus.CONFLICT:
            raise ThreadNetworkActiveError
        if response.status != HTTPStatus.OK:
            raise OTBRError(f"unexpected http status {response.status}")

    async def create_pending_dataset(self, dataset: PendingDataSet) -> None:
        """Create pending operational dataset.

        The passed in PendingDataSet does not need to be fully populated, any fields
        not set will be automatically set by the open thread border router.
        Raises if the http status is 400 or higher or if the response is invalid.
        """
        response = await self._session.put(
            f"{self._url}/node/dataset/pending",
            json=dataset.as_json(),
            timeout=aiohttp.ClientTimeout(total=self._timeout),
        )

        if response.status == HTTPStatus.CONFLICT:
            raise ThreadNetworkActiveError
        if response.status not in (HTTPStatus.CREATED, HTTPStatus.OK):
            raise OTBRError(f"unexpected http status {response.status}")

    async def delete_pending_dataset(self) -> None:
        """Delete pending operational dataset."""
        response = await self._session.delete(
            f"{self._url}/node/dataset/pending",
            timeout=aiohttp.ClientTimeout(total=self._timeout),
        )

        if response.status == HTTPStatus.CONFLICT:
            raise ThreadNetworkActiveError
        if response.status != HTTPStatus.OK:
            raise OTBRError(f"unexpected http status {response.status}")

    async def set_active_dataset_tlvs(self, dataset: bytes) -> None:
        """Set current active operational dataset.

        Raises if the http status is 400 or higher or if the response is invalid.
        """

        response = await self._session.put(
            f"{self._url}/node/dataset/active",
            data=dataset.hex(),
            headers={"Content-Type": "text/plain"},
            timeout=aiohttp.ClientTimeout(total=10),
        )

        if response.status == HTTPStatus.CONFLICT:
            raise ThreadNetworkActiveError
        if response.status not in (HTTPStatus.CREATED, HTTPStatus.OK):
            raise OTBRError(f"unexpected http status {response.status}")

    async def set_channel(
        self, channel: int, delay: int = PENDING_DATASET_DELAY_TIMER
    ) -> None:
        """Change the channel

        The channel is changed by creating a new pending dataset based on the active
        dataset.
        """
        if not 11 <= channel <= 26:
            raise OTBRError(f"invalid channel {channel}")
        if not (dataset := await self.get_active_dataset()):
            raise OTBRError("router has no active dataset")

        if dataset.active_timestamp and dataset.active_timestamp.seconds is not None:
            dataset.active_timestamp.seconds += 1
        else:
            dataset.active_timestamp = Timestamp(False, 1, 0)
        dataset.channel = channel
        pending_dataset = PendingDataSet(active_dataset=dataset, delay=delay)

        await self.create_pending_dataset(pending_dataset)

    async def get_extended_address(self) -> bytes:
        """Get extended address (EUI-64).

        Raises if the http status is not 200 or if the response is invalid.
        """
        response = await self._session.get(
            f"{self._url}/node/ext-address",
            headers={"Accept": "application/json"},
            timeout=aiohttp.ClientTimeout(total=self._timeout),
        )

        if response.status != HTTPStatus.OK:
            raise OTBRError(f"unexpected http status {response.status}")

        try:
            return bytes.fromhex(await response.json())
        except ValueError as exc:
            raise OTBRError("unexpected API response") from exc

    async def get_coprocessor_version(self) -> str:
        """Get the coprocessor firmware version.

        Raises if the http status is not 200 or if the response is invalid.
        """

        response = await self._session.get(
            f"{self._url}/node/coprocessor/version",
            headers={"Accept": "application/json"},
            timeout=aiohttp.ClientTimeout(total=self._timeout),
        )

        if response.status != HTTPStatus.OK:
            raise OTBRError(f"unexpected http status {response.status}")

        try:
            return await response.json()
        except ValueError as exc:
            raise OTBRError("unexpected API response") from exc



"""Test data models."""

import python_otbr_api


def test_deserialize_pending_dataset():
    """Test deserializing a pending dataset."""
    assert python_otbr_api.PendingDataSet.from_json(
        {
            "ActiveDataset": {
                "NetworkName": "OpenThread HA",
            },
            "Delay": 12345,
            "PendingTimestamp": {},
        }
    ) == python_otbr_api.PendingDataSet(
        python_otbr_api.ActiveDataSet(network_name="OpenThread HA"),
        12345,
        python_otbr_api.Timestamp(),
    )
